import os
import pdb

import boto3
import pandas as pd
import pandas.io.sql as psql
from sqlalchemy import create_engine


class Aurora:
    u"""
    postgresql/redshift接続クラス
    """

    def __init__(self, is_debug=False, target='staging'):
        self.db = None
        self.target = target
        self.is_debug = is_debug

    def con(self):
        if self.target == 'production':
            if self.is_debug: print('connect production')
            self.connect_db()
        elif self.target == 'staging':
            if self.is_debug: print('connect staging')
            self.connect_staging()
        else:
            pass

    def connect_db(self, host, user, passwd, schema):
        if self.db is not None:
            return self
        connector_str = 'mysql+pymysql://{}:{}@{}/{}?charset=utf8'.format(
            user, passwd, host, schema)
        self.db = create_engine(connector_str)
        return self

    def read(self, query):
        if self.db is None:
            self.con()
        if self.is_debug: print(query)
        df = psql.read_sql(query, self.db)
        return df

    def execute(self, query):
        if self.db is None:
            self.con()
        if self.is_debug: print(query)
        try:
            with self.db.connect() as con:
                con.execute(query)
        except Exception as e:
            return str(e)
        return "success"

    def update(self, df, target_table, tmp_table):
        df.to_sql(tmp_table, self.db, if_exists='replace', chunksize=10000)
        col = ' '.join(['a.{0} = b.{0}'.format(x) for x in df.columns if x != 'id'])
        query = "update {} a inner join {} b on a.id = b.id set {}".format(target_table, tmp_table, col)
        if self.is_debug: print(query)
        with self.db.connect() as con:
            con.execute(query)

    def undersampling(self, base_query, pos, neg, unique_key, *, ratio=1.0, limit=-1, salt='saltydog', sample_id='sample_id', args={}):
        """
        DBレベルでの不均衡データアンダーサンプリング。
        posiviteデータのratio倍のnegativeデータをランダムに抽出する。

        base_query: SQL文字列でwhere句に{separater}変数を、order by句の位置に{orderby}変数を指定すること
        ex) pos = "status = 'win'"
            neg = "status = 'lose'"
            unique_key = "users.id"
            base_query = "select * from users where deleted_at is null and {separator} {orderby}"
        結果)
         ->(ok) select * from users where deleted_at is null and status = 'win'
         ->(ng) select * from users where deleted_at is null and status = 'lose' order by farm_fingerprint(users.id) limit 2000
        """
        # positiveを取得
        orderby = "order by sha2(concat({},'{}'),224) limit {}".format(
            unique_key, salt, str(limit)) if limit > 0 else ''
        query = base_query.format(separater=pos, orderby=orderby, **args)
        positive = self.read(query)
        pos_num = len(positive)
        if self.is_debug:
            print('positive num:', pos_num)

        # negativeをpositiveのrate倍取得
        limit = round(pos_num * ratio)
        orderby = "order by sha2(concat({},'{}'),224) limit {}".format(unique_key, salt, str(limit))
        query = base_query.format(separater=neg, orderby=orderby, **args)
        negative = self.read(query)
        print('negative num:', str(len(negative)))

        df = pd.concat([negative, positive], ignore_index=True)
        df[sample_id] = df.index
        return df

    def nosampling(self, base_query, condition=' 1=1 '):
        query = base_query.format(separater=condition, orderby='')
        df = self.read(query)
        if self.is_debug:
            print('num:', len(df))
        return df



import time
import math
import os
import pdb
import pickle
import re
from collections import Counter
from datetime import datetime
import subprocess
import concurrent.futures
from datetime import datetime
from shutil import move
import glob

import pandas as pd
import numpy as np
import pytz
from sklearn.model_selection import train_test_split
import itertools
import boto3


class EtlHelper:
    u"""
    データ分析ETLツール
    """

    def __init__(self, use_cache=True, is_debug=True):
        self.is_debug = is_debug
        self.use_cache = use_cache

    def dump(self, obj, filename):
        with open(filename, 'wb') as f:
            pickle.dump(obj, f, protocol=4)

    def load(self, filename):
        with open(filename, 'rb') as f:
            return pickle.load(f)

    def s3cp(self, src, dest, s3_region=None):
        opt = '--region={}'.format(s3_region) if s3_region is not None else ''
        print(['aws', opt, 's3', 'cp', src, dest])
        subprocess.call(['aws', opt, 's3', 'cp', src, dest])

    def s3sync(self, src, dest, s3_region=None):
        opt = '--region={}'.format(s3_region) if s3_region is not None else ''
        print(subprocess.call(['aws', opt, 's3', 'sync', src, dest]))

    def read_ssm(self, key, region=None):
        #TODO: regionを反映
        client = boto3.client('ssm')
        response = client.get_parameter(
            Name=key,
            WithDecryption=True,
        )
        return response

    # rekognition用のmanifestファイルを作成する
    def make_manifest(self, df, filename):
        dfx = df[df['tvt'] == 'train'].copy()
        dfx['source-ref'] = dfx['image_url']
        dfx[['source-ref']].to_json(filename, orient='records', force_ascii=False, lines=True)

    def display(self, line, *x):
        min_rows = pd.options.display.min_rows
        max_rows = pd.options.display.max_rows
        pd.options.display.min_rows = line
        pd.options.display.max_rows = line
        display(*x)
        pd.options.display.min_rows = min_rows
        pd.options.display.max_rows = max_rows

    def category2id(self, categories):
        tag_idx = sorted(set(categories))
        tag_dic = dict(zip(tag_idx, range(len(tag_idx))))
        return categories.map(tag_dic), tag_idx

    def adjust_frame(self, df, frame, drop=True):
        """
        空フレームを呼び出してカラムをtrain時の構造に一致させる(欠損しているダミーフィールドを補完)
        """
        columns_diff = set(frame.columns) - set(df.columns)
        for x in columns_diff:
            df[x] = frame[x]
        for x in frame:
            df[x] = df[x].astype(frame[x].dtype)

        if drop:
            columns_diff = set(df.columns) - set(frame.columns)
            for x in columns_diff:
                df.drop(x, axis=1, inplace=True)
        return df

    def execute(self, method, filename, **kwargs):
        if self.use_cache and os.path.isfile(filename):
            if self.is_debug: print('using cache.')
            return self.load(filename)

        obj = method(**kwargs) if len(kwargs) > 0 else method()

        if self.use_cache and filename != "":
            if self.is_debug: print('saving cache.')
            d = os.path.dirname(filename)
            if not os.path.isdir(d): os.makedirs(d)
            self.dump(obj, filename)
        return obj

    def parallel(self, future_method, args, *,
                 chunk=2000, workers=-1, executor='thread', cache=None, cache_dir=None, limit=1000000000, wait=0):
        """
        大規模データを使って外部APIを叩く場合のチャンク分割+並列処理
        * 1チャンクごとにThreadまたはProcessでfuture_methodを並列実行
        * 1チャンクごとにキャッシュデータを保存(途中再開可能)可能
        * Google APIのように1分ごとの連続アクセス数制限のあるAPIのためにチャンク単位でウエイト設定可能

        ARGS:
        future_method: 並列実行する関数(引数は配列1つであること)
        args: future_methodに渡す引数の配列
        chunk: 1チャンクごとの処理数
        workers: 1チャンクごとの並列数
        executor: thread/process/debugのいずれかの文字列
        cache: キャッシュする場合のファイル名
        cache_dir: chankごとに別ファイルとなるキャッシュ。
        limit: argsの実行行数(動作テストなどで全部実行しない場合に利用する)
        wait: チャンクごとの最小実行時間(sec)。実際の実行時間がwait値以下の場合、sleepを掛ける
        """
        start = 0
        stop = min(len(args), limit)
        digit = len(str(stop))
        results = []

        # キャッシュ呼び出し
        if cache_dir != None:
            files = sorted(glob.glob(os.path.join(cache_dir, '*.cache')))
            if len(files) > 0:
                # 最後のファイルは壊れている可能性があるのでひとつ前から
                start = (len(files) - 1) * chunk

        elif cache != None and os.path.isfile(cache):
            results = self.load(cache)
            start = len(results)

        if executor == 'thread':
            exec = concurrent.futures.ThreadPoolExecutor
            workers = os.cpu_count() * 5 if workers == -1 else workers
        else:
            exec = concurrent.futures.ProcessPoolExecutor
            workers = os.cpu_count() if workers == -1 else workers

        # チャンク分割処理
        for suffix, i in enumerate(range(start, stop, chunk)):
            if cache_dir != None:
                results = []

            start_time = datetime.now(pytz.timezone('Asia/Tokyo'))
            print(f'{i:0{digit}}-', end='')
            t = min(i + chunk, stop)
            target = args[i:t]

            # 並列処理
            futures = []
            with exec(max_workers=workers) as exe:
                for n, arg in enumerate(target):
                    if executor == 'debug':
                        futures.append(future_method(arg))
                    else:
                        futures.append(exe.submit(future_method, arg))
            ret = [x.result() if hasattr(x, 'result') else x for x in futures]
            results += ret

            # 結果キャッシュ処理
            if cache_dir != None:
                d = os.path.dirname(cache_dir)
                if not os.path.isdir(d): os.makedirs(d)
                self.dump(results, os.path.join(cache_dir, f'{i:07}.cache'))
            elif cache != None:
                d = os.path.dirname(cache)
                if not os.path.isdir(d): os.makedirs(d)
                if os.path.exists(cache):
                    move(cache, cache + '.bak')
                self.dump(results, cache)

            # チャンク終了報告
            end_time = datetime.now(pytz.timezone('Asia/Tokyo'))
            elapse = (end_time - start_time).total_seconds()
            minutes = '{:.0f}m '.format(elapse // 60) if elapse // 60 != 0 else ''
            sec = elapse % 60
            z = t - 1
            print(f'{z:0{digit}}: {minutes}{sec:.3f}s', end='')

            # ウエイト処理
            if wait > 0:
                # wait秒経過までsleepする
                remain = wait - elapse
                if remain > 0:
                    print(' sleep:', int(remain), 'sec', end='')
                    time.sleep(remain)

            print('')

        # チャンク別キャッシュの場合、全体を呼び戻す
        if cache_dir != None:
            files = sorted(glob.glob(os.path.join(cache_dir, '*.cache')))
            results = []
            for x in files:
                results += self.load(x)
        return results

    def train_valid_test_split(self, df, train_size, valid_size=None, stratify=None):
        if valid_size is None:
            valid_size = (1 - train_size) / 2
        remain_valid_size = valid_size / (1 - train_size)
        train, remain = train_test_split(df.copy(), train_size=train_size,
                                         stratify=df[stratify] if stratify in df else None, random_state=13)
        valid, test = train_test_split(remain.copy(), train_size=remain_valid_size,
                                       stratify=remain[stratify] if stratify in df else None, random_state=13)
        return train, valid, test

    def timeline_split(self, df, train_size, valid_size=None, stratify=None, orderby=None):
        if valid_size is None:
            valid_size = (1 - train_size) / 2

        if stratify is not None:
            train, valid, test = [], [], []
            for name, group in df.groupby(stratify):
                tr, va, te = self._split(group, train_size, valid_size, orderby)
                train.append(tr)
                valid.append(va)
                test.append(te)
            train = pd.concat(train)
            valid = pd.concat(valid)
            test = pd.concat(test)
        else:
            train, valid, test = self._split(df, train_size, valid_size, orderby)
        return train, valid, test

    def _split(self, dfx, train_size, valid_size, orderby):
        if orderby is not None: df = dfx.sort_values(orderby)
        maxlen = len(df)
        train_max = math.ceil(maxlen * train_size)
        valid_max = train_max + math.ceil(maxlen * valid_size)
        train = df[0:train_max].sample(frac=1, random_state=0)
        valid = df[train_max:valid_max].sample(frac=1, random_state=0)
        test = df[valid_max:].sample(frac=1, random_state=0)
        return train, valid, test

    def merge_tvt(self, merge_column, train, valid, test):
        train[merge_column] = 'train'
        valid[merge_column] = 'valid'
        test[merge_column] = 'test'
        df = pd.concat([train, valid, test])
        df = df.reset_index(drop=True)
        return df

    def show_posneg_matrix(self, df, dependant, merge_column):
        stat = pd.DataFrame(
            {
                'all': Counter(df[dependant]),
                'train': Counter(df[df[merge_column] == 'train'][dependant]),
                'valid': Counter(df[df[merge_column] == 'valid'][dependant]),
                'test': Counter(df[df[merge_column] == 'test'][dependant]),
            }
        ).T.fillna(0).astype(int)
        print('pos_neg matrix')
        display(stat)

    def freq(self, data, class_width=None):
        data = np.asarray(data)
        if class_width is None:
            class_size = int(np.log2(data.size).round()) + 1
            class_width = round((data.max() - data.min()) / class_size)

        bins = np.arange(0, data.max() + class_width + 1, class_width)
        hist = np.histogram(data, bins)[0]
        cumsum = hist.cumsum()

        return pd.DataFrame(
            {
             '度数': hist,
             '累積度数': cumsum,
             '相対度数': hist / cumsum[-1],
             '累積相対度数': cumsum / cumsum[-1]
             },
            index=pd.Index(
                [f'{bins[i]}-{bins[i+1]}'
                 for i in range(hist.size)],
                name='Class')
        )

    def varplot(self, df):
        '''
        DataFrame内の変数をよい感じに並べて可視化。
        汎用化するには、描画が重くなるunique(つまりID)な列への対応が必要。
        '''
        import matplotlib.dates as mdates
        import matplotlib.pyplot as plt
        import matplotlib.ticker as ticker
        import seaborn as sns

        max_col = 5
        max_row = math.ceil(len(df.columns) / max_col)
        fig, ax = plt.subplots(max_row, max_col, figsize=[20, 3.7 * max_row])

        r = 0; c = 0
        for name in df:
            if df[name].dtype == 'datetime64[ns]':
                x = df.set_index(name, drop=False)[name].dropna().resample('D').count()
                ax[r][c].plot(x)
                ax[r][c].set_xlabel(name)
                ax[r][c].get_xaxis().set_major_formatter(mdates.DateFormatter('%Y/%m/%d'))
                ax[r][c].get_xaxis().set_major_locator(ticker.MaxNLocator(integer=True, nbins=3))
            elif name.endswith('_id'):
                ax[r][c].text(0.5, 0.5, name + ' is ID', horizontalalignment='center', verticalalignment='center')
                ax[r][c].set_xlabel(name)
            elif df[name].dtype == 'object':
                sns.countplot(df[name].fillna(False), ax=ax[r][c])
            else:
                sns.distplot(df[name].dropna(), ax=ax[r][c])
            # 位置情報のインクリメント
            (c, r) = (0, r + 1) if (c == max_col - 1) else (c + 1, r)
        return fig


class Pipeline:
    """
    キャッシュ機能付き連続実行
    辞書の配列で実行する処理群を記載
    辞書内のキーとして、method, cachename, args, kwargs を指定可能
      name: 処理名(必須)
      method: 実行する関数ポインタ(必須)
      return: 戻り値の変数名
      cache: cacheファイル名。省略するとここではキャッシュを作らない
      args: methodの引数配列
      kwargs: methodの名前付き引数辞書
    cacheが指定されていてそのファイルが存在した場合、それ以前の処理はスキップされる
    args, kwargsに '#'で囲まれた変数名を指定すると、それ以前の処理で生成された戻り値を利用できる
    ---
    example:
    steps = [
            {
                'name': 'transform',
                'method': self.loader.transform,
                'cache': True,
                'return': ('df'),
                'args': [df],
            },
            {
                'name': 'preprocess',
                'method': self.loader.preprocess,
                'return': ('df', 'categories'),
                'args': ['#df#'],
            },
    ]
    """

    def __init__(self, steps, use_cache=True,
                 is_debug=True, cache_dir='./cache',
                 returns=[],
                 ):
        self.steps = steps
        self.is_debug = is_debug
        self.use_cache = use_cache
        self.cache = cache_dir + '{}.cache'
        self.returns = returns

        self.vars = re.compile('^\#.*\#$')
        self.helper = Helper()

    def run(self, breakpoint=None):
        steps = self.steps
        ret = {}

        # キャッシュ存在チェックと呼び出し
        start = 0
        for i, x in self.enumerate_reversed(steps):
            cache = self.cache.format(x['name'])
            if self.use_cache and 'cache' in x and x['cache'] and os.path.exists(cache):
                # キャッシュがあった場合呼び出し
                print('recent cache: ', cache)
                ret = self.helper.load(cache)
                start = i + 1
                break;

        # キャッシュを呼び出したところ以降から処理を連続実行
        breakpoint = max(min(len(steps), breakpoint), start) if breakpoint is not None else len(steps)
        for x in steps[start:breakpoint]:

            print('start:', x['name'], '-------------------')
            start = datetime.now(pytz.timezone('Asia/Tokyo'))

            args = x['args'] if 'args' in x and len(x['args']) > 0 else []
            args = [self.parse_arg(y, ret) for y in args]

            kwargs = x['kwargs'] if 'kwargs' in x and len(x['kwargs']) > 0 else {}
            kwargs = {k: self.parse_arg(y, ret) for k, y in kwargs.items()}

            obj = x['method'](*args, **kwargs)
            ret = self.parse_return(x, obj, ret)

            if self.use_cache and 'cache' in x and x['cache']:
                cache = self.cache.format(x['name'])
                d = os.path.dirname(cache)
                if not os.path.isdir(d): os.makedirs(d)
                self.helper.dump({k: ret[k] for k in self.returns if k in ret}, cache)

            end = datetime.now(pytz.timezone('Asia/Tokyo'))
            self.print(start, end, 'done : {} ----------'.format(x['name']))

        return tuple([ret[x] if x in ret else None for x in self.returns]) if len(self.returns) > 1 else ret[self.returns[0]]

    def enumerate_reversed(self, lyst):
        length = len(lyst) - 1
        for index, value in enumerate(reversed(lyst)):
            index = length - index
            yield index, value

    def parse_return(self, x, obj, ret):
        if 'return' not in x:
            return ret
        if type(x['return']) is str:
            ret[x['return']] = obj
            return ret
        if len(x['return']) == 1:
            ret[x['return'][0]] = obj
            return ret
        for i, y in enumerate(x['return']):
            ret[y] = obj[i]
        return ret

    def parse_arg(self, arg, ret):
        if type(arg) is str and self.vars.match(arg):
            varname = arg.replace("#", '')
            return ret[varname]
        return arg

    def print(self, start: datetime, end: datetime, name=None):
        elapse = (end - start).total_seconds()
        min = '{:.0f}m '.format(elapse // 60) if elapse // 60 != 0 else ''
        sec = elapse % 60
        print('{}: {}{:.3f}s'.format(name, min, sec))


class StopWatch:
    """
    実行時間計測
    """

    def __init__(self):
        self.start = datetime.now(pytz.timezone('Asia/Tokyo'))
        print('start:', self.start.strftime('%Y/%m/%d %H:%M:%S %Z'))
        self.pre = self.start
        self.i = 0

    def stop(self):
        end = datetime.now(pytz.timezone('Asia/Tokyo'))
        if self.pre != self.start:
            self.split('last')
        self.print(self.start, end, 'total')
        print("done:", end.strftime('%Y/%m/%d %H:%M:%S %Z'))

    def split(self, name='elapse'):
        self.i += 1
        end = datetime.now(pytz.timezone('Asia/Tokyo'))
        self.print(self.pre, end, "{:02d}_{}".format(self.i, name))
        self.pre = end

    def print(self, start: datetime, end: datetime, name=None):
        elapse = (end - start).total_seconds()
        min = '{:.0f}m '.format(elapse // 60) if elapse // 60 != 0 else ''
        sec = elapse % 60
        print('{}: {}{:.3f}s'.format(name, min, sec))



from .aurora import *
from .awstool import *
from .bigquery import *
from .bq_kintone import *
from .config_abc import *
from .elasticcache import *
from .etltool import *
from .evaluator import *
from .kintone import *
from .parallelget import *

__copyright__ = 'Copyright (C) 2023 Takemi Ohama'
__VERSION__ = '0.2.1'
__license__ = 'MIT'
__author__ = 'Takemi Ohama'
__author_email__ = 'takmei.ohama@gmail.com'
__url__ = 'https://github.com/takemi-ohama/tmllib'



import requests
import numpy as np
import math
import json
from .etltool import EtlHelper

class Kintone:
    """
    kintone API
    zenkPythonをベースにGPT-4によるリファクタリングと一部機能追加したライブラリです。
    https://github.com/zenk-github/pytone
    """

    BASE_URL_TEMPLATE = 'https://{}.cybozu.com/k/v1/{}'

    def __init__(self, api_token, domain, app):
        self.api_token = api_token
        self.base_url = self.BASE_URL_TEMPLATE.format(domain, '{}')
        self.app = app
        self.headers = {
            "X-Cybozu-API-Token": self.api_token,
            'Content-Type': 'application/json'
        }
        self.property, self.fields = self._get_property()
        self.helper = EtlHelper()

    def select_all(self, where=None, fields=None, hard_limit=None):
        params = {
            'app': self.app,
            'query': '',
            'totalCount': True,
        }
        if fields is not None:
            params['fields'] = list(set(fields + ['$id', '$revision']))

        records = self._fetch_records_in_batches(params, where, hard_limit)
        records = self._format_records(records)
        return records

    def _request_kintone(self, method, endpoint, json_data=None):
        url = self.base_url.format(endpoint)
        try:
            # GETリクエストではparamsを使用、それ以外ではjsonを使用
            if method == 'GET':
                response = requests.request(method, url, params=json_data, headers={'X-Cybozu-API-Token': self.api_token})
            else:
                response = requests.request(method, url, json=json_data, headers=self.headers)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            raise Exception(f"Request failed: {str(e)}")

    def _get_property(self):
        params = {'app': self.app, 'lang': 'default'}
        property = self._request_kintone('GET', 'app/form/fields.json', json_data=params)['properties']
        fields = {y['label']: y for y in property.values()}
        fields |= {
            k: {'type': 'NUMBER', 'code': k, 'label': k, 'required': 'True'}
            for k in ('$id', '$revision')
        }
        return property, fields

    def _fetch_records_in_batches(self, params, where, hard_limit):
        last_rec_id = '0'
        record_count = 0
        all_records = []

        while True:
            query_str = f'($id > {last_rec_id})'
            if where is not None:
                query_str += f' and ({where})'
            params['query'] = f'{query_str} order by $id asc limit 500'

            response = self._request_kintone('GET', 'records.json', json_data=params)
            total_count = int(response['totalCount'])
            record_count += len(response['records'])

            if total_count == 0:
                break

            last_rec_id = response['records'][-1]['$id']['value']
            all_records.append(response['records'])

            if total_count <= 500:
                break
            if record_count >= hard_limit:
                break

        return [record for batch in all_records for record in batch]

    def _format_records(self, records):
        return [
            {self.property[field_code]['label'] if field_code not in (
                '$id', '$revision') else field_code: self._format_field(value) for field_code, value in record.items()}
            for record in records
        ]

    def _format_field(self, field_data):
        field_type = field_data['type']
        field_value = field_data['value']

        if field_type == 'NUMBER' and field_value is not None and field_value != "":
            return self._convert_to_number(field_value)
        elif field_type == 'SUBTABLE':
            return self._format_subtable(field_value)
        else:
            return field_value

    def update(self, records: list):
        """kintone rest apiの制限(updateは1回100件)に従って分割送信"""
        cnt = math.ceil(len(records) / 100)
        chunk = list(np.array_split(records, cnt))
        # 並列処理を無効化して順次処理に変更（404エラー回避）
        for c in chunk:
            self._update_chunk(c)
        return

    def _update_chunk(self,params):
        # paramsを正しいkintone API形式に変換
        # 既に'record'キーが存在する場合はそのまま使用、ない場合は追加
        records = []
        for param in params:
            param_copy = param.copy()
            if 'record' in param_copy:
                # 既にrecordキーがある場合はそのまま使用
                records.append(param_copy)
            else:
                # recordキーがない場合は追加
                record_id = param_copy.pop('id')
                records.append({
                    'id': record_id,
                    'record': param_copy
                })
        data = {'app': int(self.app), 'records': records}
        response = self._request_kintone('PUT', 'records.json', json_data=data)
        return response

    @staticmethod
    def _convert_to_number(value):
        try:
            return int(value)
        except ValueError:
            return float(value)

    @staticmethod
    def _format_subtable(subtable_value):
        formatted_subtable = []

        for sub_rec in subtable_value:
            subtable_record = {'id': sub_rec['id']}

            for sub_field_code, sub_value in sub_rec['value'].items():
                if sub_value['type'] == 'NUMBER' and sub_value['value'] is not None:
                    sub_value['value'] = Kintone._convert_to_number(sub_value['value'])

                subtable_record[sub_field_code] = sub_value['value']
            formatted_subtable.append(subtable_record)

        return formatted_subtable



import pdb
import pandas as pd
import redis

class Redis:
    u"""
    Redis接続クラス
    """

    def __init__(self, host = '', db = 0, is_debug=False):
        self.is_debug = is_debug
        self.host = host
        self.__conn = None
        self.__db = db

    @property
    def conn(self):
        if self.__conn is not None:
            return self.__conn
        self.__conn = redis.Redis(host = self.host, port = 6379, db = self.db)
        return self.__conn

    @property
    def db(self):
        return self.__db
        
    @db.setter
    def db(self, value):
        if value != '':
            self.__db = value
            self.__conn = redis.Redis(host = self.host, port = 6379, db = self.__db)
        
    def set(self, key, value):
        self.conn.set(key.value)
        return df
    
    def get(self, key):
        self.conn.get(key.value)
        return df
    
    def mset(self, data):
        return self.conn.mset(data)
    
    def mget(self,keys):
        return self.conn.mget(keys)
    
    def mget_map(self,keys):
         return dict(zip(keys, self.conn.mget(keys)))
    



import json
import pdb

import pandas as pd
from google.oauth2 import service_account
import boto3
import os
from google.cloud import bigquery
from .config_abc import BaseConfig


class BigQuery:
    u"""
    Google BigQuery接続クラス
    """

    def __init__(self, conf: BaseConfig):
        self.conf = conf
        self.account_type = self.conf.account_type
        self.project_id = self.conf.project_id if hasattr(self.conf, 'project_id') else None
        self.json_key = self.conf.json_key if hasattr(self.conf, 'json_key') else None
        self.region = self.conf.aws_region if hasattr(self.conf, 'aws_region') else None
        self._cred = None
        self.client = None

    def read_gbq(self, query, args={}):
        query = query.format(**args)
        if self.conf.is_debug: print(query)
        df = pd.read_gbq(query, project_id=self.project_id, dialect='standard', credentials=self.cred())
        return df

    def write_gbq(self, df, tablename, table_schema=None, location=None, if_exists='replace'):
        df.to_gbq(tablename, project_id=self.project_id, if_exists=if_exists,
                  credentials=self.cred(), table_schema=table_schema, location=location)
        return df

    def cred(self):
        if self._cred is not None:
            return self._cred
        if self.account_type == 'file':
            self._cred = service_account.Credentials.from_service_account_file(self.json_key)
        elif self.account_type == 'ssm':
            session = boto3.Session(profile_name=self.conf.aws_profile, region_name=self.conf.aws_region)
            param_json = session.client('ssm').get_parameter(
                Name=self.json_key, WithDecryption=True)['Parameter']['Value']
            json_key = json.loads(param_json)
            self._cred = service_account.Credentials.from_service_account_info(json_key)
        elif self.account_type == 'env':
            self._cred = service_account.Credentials.from_service_account_file(
                os.getenv('GOOGLE_APPLICATION_CREDENTIALS'))
        self.client = bigquery.Client(project=self.project_id, credentials=self._cred)
        return self._cred

    def jsoncolumn_to_df(self, data, prefix=None):
        lst = data.values.tolist()
        json_str = '[' + ','.join(lst) + ']'
        df_s = pd.read_json(json_str)
        if prefix != None:
            prefix = prefix + '_'
            df_s.columns = [prefix + x for x in df_s.columns]
        return df_s

    def undersampling(self, base_query, pos, neg, unique_key, ratio=1.0, limit=-1, args={}):
        """
        DBレベルでの不均衡データアンダーサンプリング。posのデータ数のratio倍のnegデータをランダムに抽出する。
        ここではnegativeを5-10倍出してコード側でSMOTEENNを掛けるなどの操作を推奨

        base_query: SQL文字列でwhere句に{separater}変数を、order by句の位置に{orderby}変数を指定すること
        ex) pos = "status = 'win'"
            neg = "status = 'lose'"
            unique_key = "users.id"
            base_query = "select * from users where deleted_at is null and {separator} {orderby}"

         ->(ok) select * from users where deleted_at is null and status = 'win'
         ->(ng) select * from users where deleted_at is null and status = 'lose' order by farm_fingerprint(users.id) limit 2000
        """
        # positiveを取得
        orderby = 'order by farm_fingerprint({}) limit {}'.format(unique_key, str(limit)) if limit > 0 else ''
        query = base_query.format(separater=pos, orderby=orderby, **args)
        positive = self.read_gbq(query)
        pos_num = len(positive)
        if self.conf.is_debug:
            print('positive num:', pos_num)

        # negativeをpositiveのrate倍取得
        limit = round(pos_num * ratio)
        orderby = 'order by farm_fingerprint({}) limit {}'.format(unique_key, str(limit))
        query = base_query.format(separater=neg, orderby=orderby, **args)
        negative = self.read_gbq(query)
        print('negative num:', str(len(negative)))

        df = pd.concat([negative, positive], ignore_index=True)
        return df

    def nosampling(self, base_query, condition=' 1=1 '):
        query = base_query.format(separater=condition, orderby='')
        df = self.read_gbq(query)
        if self.conf.is_debug:
            print('num:', len(df))
        return df

    def upload_csv(self, filename, tablename, **kwarg):
        df = pd.read_csv(filename, **kwarg)
        self.write_gbq(df, tablename)

    def query_with_noreturn(self, sql):
        self.cred()
        if self.conf.is_debug: print(sql)
        self.client.query(sql).result()

    def query(self, sql):
        self.cred()
        if self.conf.is_debug: print(sql)
        return self.client.query(sql).to_dataframe()



import os
import hashlib
import pdb
import requests
from .etltool import EtlHelper

class Downloader:

    def __init__(self,cache_dir='/home/sagemaker-user/SageMaker/storage/image_cache'):

        #with共通で一つの画像キャッシュを持つ
        self.cache_dir = cache_dir
        self.helper = EtlHelper()

    def download_files(self, urls):
        files = self.helper.parallel(self.download, urls)
        return files

    def download(self, url):
        if url is None:
            return None

        _, ext = os.path.splitext(url)
        basename = hashlib.sha256(url.encode('utf-8')).hexdigest() + ext
        subdir = basename[:3]
        filename = os.path.join(self.cache_dir, subdir, basename)
        os.makedirs(os.path.join(self.cache_dir, subdir), exist_ok=True)

        if os.path.exists(filename):
            if os.path.getsize(filename) == 0:
                return '#forbidden'
            return filename

        try:
            r = requests.get(url)
            if r.status_code == requests.codes.forbidden:
                with open(filename, 'w') as f:
                    f.write('')
                return '#forbidden'
            r.raise_for_status()
            with open(filename, 'wb') as f:
                f.write(r.content)
        except Exception as e:
            print(e)
            raise e
        return filename



import pandas as pd
import sklearn.metrics as mt
import traceback


class Evaluator:
    u"""
    トレーニング関連ライブラリ
    """

    def __init__(self, train_X=None, train_y=None, test_X=None, test_y=None):
        self.train_X, self.train_y, self.test_X, self.test_y = train_X, train_y, test_X, test_y
        self.clf = None;

    def show_importance(self, feature_importance, test_X, pred_y=None, num=20):
        # 特徴量名と重要度を含むデータフレームを作成
        feature_info = pd.DataFrame({'name': test_X.columns, 'importances': feature_importance})
        feature_info.set_index('name', inplace=True)
        feature_info['type'] = test_X.dtypes

        # 重要度が0でない項目を抽出
        non_zero_importance = feature_info[feature_info['importances'] != 0].copy()
        zero_importance = feature_info[feature_info['importances'] == 0].copy()

        # 特徴量の詳細情報を表示
        print("重要度リスト-----------------------------------")
        # self.printfull(non_zero_importance)

        # 重要度が0の項目を表示
        print("重要度0の特徴量---------------------------------")
        # self.printfull(zero_importance)

        # 重要度に基づいてトップとワーストの特徴量を表示
        print(f"トップ{num}------------------------------")
        non_zero_importance.sort_values("importances", ascending=False, inplace=True)
        self.print_topx(non_zero_importance, num)

    def show_importances(self, clf, test_X=None, test_y=None, pred_y=None):
        self.show_importance(clf.feature_importances_, test_X, pred_y)

    def printfull(self, x):
        max_rows = pd.get_option('display.max_rows')
        pd.set_option('display.max_rows', None)
        try:
            display(x)
        except NameError:
            print(x)
        pd.set_option('display.max_rows', max_rows)

    def print_topx(self, x, num=20):
        with pd.option_context('display.max_rows', num * 2):
            try:
                display(x.head(num))
            except NameError:
                print(x.head(num))

    def show_valuation(self, test_y, pred_y, prob_y=None):
        print("...valuation...")
        report = mt.classification_report(test_y, pred_y, output_dict=True)
        df = pd.DataFrame(report).T
        df['support'] = df['support'].astype('int')
        try:
            display(df)
        except NameError:
            print(df)
        print('accuracy_score: {0:.3f}'.format(mt.accuracy_score(test_y, pred_y)))
        if prob_y is not None:
            try:
                print('average_precision: {0:.3f}'.format(mt.average_precision_score(test_y, prob_y)))
                print('roc_auc: {0:.3f}'.format(mt.roc_auc_score(test_y, prob_y)))
            except Exception as e:
                mes = traceback.format_exception_only(type(e), e)
                if 'multi' not in mes[0]:
                    raise e
        print('MCC: {0:.3f}'.format(mt.matthews_corrcoef(test_y, pred_y)))

    def show_roc(self, test_y, prob_y):
        roc_auc = mt.roc_auc_score(test_y, prob_y)
        fpr, tpr, thresholds = mt.roc_curve(test_y, prob_y)
        self.plot_roc(fpr, tpr, roc_auc)

    def plot_roc(self, fpr, tpr, roc_auc, title='ROC curve'):
        import matplotlib.pyplot as plt
        plt.rcParams['font.family'] = 'IPAexGothic'
        plt.figure()
        lw = 2
        plt.plot(fpr, tpr, color='darkorange',
                 lw=lw, label=f'{title} (area = %0.2f)' % roc_auc)
        plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title(title)
        plt.legend(loc="lower right")
        plt.show()

    def show_roc_multiclass(self, test_ys, prob_ys):
        for name in prob_ys:
            test_y = test_ys[name]
            prob_y = prob_ys[name]
            roc_auc = mt.roc_auc_score(test_y, prob_y)
            fpr, tpr, thresholds = mt.roc_curve(test_y, prob_y)
            self.plot_roc(fpr, tpr, roc_auc, title=name)



from enum import Enum
from pydantic import BaseModel
from typing import Optional


class AccountType(str, Enum):
    ssm = 'ssm'
    file = 'file'
    env = 'env'


class BaseConfig(BaseModel):
    account_type: AccountType
    json_key: str
    project_id: str
    is_debug: bool
    aws_region: Optional[str]
    aws_profile: Optional[str]
    gbq_location: Optional[str]
    app_list: Optional[str]




import json
import boto3
import pandas as pd
import hashlib
from google.api_core.exceptions import NotFound, Conflict
from .config_abc import BaseConfig
from .etltool import EtlHelper
from .bigquery import BigQuery
from .kintone import Kintone
import datetime


class BQKintone:

    def __init__(self, conf: BaseConfig, tables):
        self.conf = conf
        session = boto3.Session(region_name=self.conf.aws_region)
        param_json = session.client('ssm').get_parameter(
            Name=self.conf.app_list, WithDecryption=True)['Parameter']['Value']
        self.apps = json.loads(param_json)
        self.db = BigQuery(conf)
        self.helper = EtlHelper()
        self.schema_type = 'type'

        # フィールド名変換辞書。結局ほぼ全部_に変換したので、setで十分かも。
        self.ng_fields = str.maketrans({
            '!': '_', '"': '_', '$': '_', '(': '_', ')': '_', '*': '_', ',': '_',
            '.': '_', '/': '_', ';': '_', '?': '_', '@': '_', '[': '_', '\\': '_',
            ']': '_', '^': '_', '`': '_', '{': '_', '}': '_', '~': '_', '【': '_',
            '】': '_', '−': '_', '（': '_', '）': '_', '・': '_', '、': '_',
            '\ufeff': '', '※': '_', '。': '_', '：': '_', '　': '_', '→': '_',
        })

        # [{kintoneアプリ名: bigquertテーブル名},{...}...]
        self.tables = tables

    def run(self):
        """idで全テーブル取得"""
        for k, v in self.tables.items():
            last_id = self._get_last_id(v)
            where = f"$id > {last_id}"
            self.etl(v, k, where)

    def insert_as_updated(self, appname):
        tablename = self.tables[appname]
        last_updated_at = self._get_last_updated_at(tablename)
        where = f'更新日時 > "{last_updated_at}"'
        print("where", where)
        self.etl(tablename, appname, where)

    def insert_as_id(self, appname):
        tablename = self.tables[appname]
        last_id = self._get_last_id(tablename)
        where = f"$id > {last_id}"
        self.etl(tablename, appname, where)

    def update_kintone(
        self, app_name,
        id: pd.Series,
        values: pd.DataFrame,
        field_codes=None
    ):
        """kintone側の情報を更新する"""
        # fields_codeとpandas.columnsの対応表を作成
        col_dic = (
            {x: y for x, y in zip(values.columns, field_codes)}
            if field_codes is not None else {x: x for x in values.columns}
        )

        # kintone apiのrecordsオブジェクトを作成
        # ref: https://cybozu.dev/ja/kintone/docs/rest-api/records/update-records/
        t = values.T
        records = [
            {'id': int(x),
             'record': {col_dic[y]: {'value': x} for x, y in zip(t[i], t[i].index)}
             }
            for i, x in enumerate(id)
        ]

        info = self.apps[app_name]
        app = Kintone(info['api_token'], info['sub_domain'], info['app_id'])
        res = app.update(records)
        return res

    def etl(self, tablename, appname, where):
        '''
        kintoneからbigqueryへのデータ転送処理
        Parquetの制約が厳しいので、日本語フィールド名を一度md5に変換してテーブルを作成。
        その後alter table rename columnで日本語フィールド名に戻す。
        '''

        # kintoneからデータを取得
        df, fields = self._select(app_name=appname, where=where, limit=10000)
        if len(df) == 0:
            print(appname, 'to', tablename, ':', len(df))
            return

        # フィールド名をMD5に変換し、辞書を保存。
        col_utf8, fields_md5 = self._hash_fields(df, fields)

        # スキーマを生成してframeを型変換
        schema = self._create_schema(df, fields_md5)
        df = self._adjust_type(df, schema)

        # bigquerryにtmp書き込み
        self.db.write_gbq(df, tablename=tablename + '_tmp2', table_schema=schema)

        # numeric型と日付型を型変換
        self._change_type_bq(tablename, fields_md5)

        # フィールド名をクリーニング
        col_utf8 = self._clean_fieldname(col_utf8)

        # bigqueryのフィールド名を日本語に変更
        sql = self._rename_sql(tablename + '_tmp', df.columns, col_utf8)
        self.db.query_with_noreturn(sql)

        # insert intoに渡すフィールド一覧を作成
        tmp_fields = self._get_fieldnames(f'{tablename}_tmp')
        fields = "`" + "`, `".join(tmp_fields['column_name'].to_list()) + "`"

        # 初回実行用にcreate table
        self._create_original(tablename, f'{tablename}_tmp')

        # 存在しないフィールドを検知したときはalter table add columnを実行
        self._add_columns(tablename, tmp_fields)

        # tmpからオリジナルにコピー
        self._insert_original(tablename, fields)

        # id,revisionが同じレコードは削除
        self._drop_duplicated(tablename)

        # tmpを削除
        sql = f"drop table {tablename}_tmp; drop table {tablename}_tmp2"
        self.db.query_with_noreturn(sql)

        # report
        print(appname, 'to', tablename, ':', len(df))

    def _drop_duplicated(self, tablename):
        """id,revisionが同じレコードは削除"""
        sql = f"""
            merge {tablename} table
            using (
                select
                    id,revision,
                    min(inserted_at) as inserted_at
                from
                    {tablename}
                group by
                    id,revision
                having count(id) > 1
            ) duplicated_origin
            on  table.id = duplicated_origin.id and table.revision = table.revision
            and table.inserted_at != duplicated_origin.inserted_at
            when matched then
                delete
            ;
        """
        self.db.query_with_noreturn(sql)

    def _insert_original(self, tablename, fields):
        """tmpテーブルからoriginalテーブルへのinsert"""
        try:
            sql = f"insert into {tablename} ({fields},`inserted_at`) select *,current_timestamp() from {tablename}_tmp"
            self.db.query_with_noreturn(sql)
        except NotFound as e:
            sql = f"create table {tablename} clone {tablename}_tmp;"
            self.db.query_with_noreturn(sql)

    def _create_original(self, original, tmptable):
            """tmpテーブルからoriginalテーブルを作成"""
            sql = f"create table if not exists {original} as select id,revision,* except(id,revision), current_timestamp() as inserted_at from {tmptable};"
            self.db.query_with_noreturn(sql)

    def _add_columns(self, tablename, tmp_fields):
        """originalにないフィールドを検出した際にフィールドをalter table add columnする"""
        org_fields = self._get_fieldnames(f'{tablename}')
        tmp_set = set(tmp_fields['column_name'].to_list())
        org_set = set(org_fields['column_name'].to_list())
        tmp_type = {x: y for x, y in zip(tmp_fields['column_name'], tmp_fields['data_type'])}
        org_set = set(org_fields['column_name'].to_list())

        # 差集合を取って新フィールドを検出
        diff = tmp_set - org_set
        if len(diff) == 0: return

        # 新フィールドの型を取得してsqlを作成
        ret = [(x, tmp_type[x]) for x in diff]
        add_col = ', '.join([f"add column `{name}` {type}" for name, type in ret])
        sql = f"alter table {tablename} {add_col}"
        self.db.query_with_noreturn(sql)

    def _get_fieldnames(self, tablename):
        """"フィールド名と型の一覧をdataframeで返す"""
        dataset, table = tablename.split('.')
        sql = f"""
            select column_name,data_type,
            from {dataset}.INFORMATION_SCHEMA.COLUMNS
            where table_name='{table}'
            order by ordinal_position
        """
        df = self.db.read_gbq(sql)
        return df

    def _hash_fields(self, df, fields):
        """parquetの制約を回避するため、一旦フィールド名を全てmd5ハッシュにする"""
        def md5(x): return hashlib.md5(x.encode()).hexdigest()
        col_utf8 = df.columns
        df.columns = df.columns.map(md5)
        fields_md5 = {md5(k): v for k, v in fields.items()}
        return col_utf8, fields_md5

    def _get_last_updated_at(self, tablename):
        """bigquery転送済みデータの最新更新日次の2時間前を返す"""
        sql = f"select max(`更新日時`) as last_updated_at from {tablename}"
        try:
            df = self.db.query(sql)
        except NotFound as e:
            return '1900-01-01'
        if len(df) == 0:
            return '1900-01-01'
        last_updated_at = df['last_updated_at'][0] + datetime.timedelta(hours=-2)
        last_updated_at = last_updated_at.strftime('%Y-%m-%dT%H:%M:%SZ')
        return last_updated_at

    def _get_last_id(self, tablename):
        """bigquery転送済みデータの最新更新idを返す"""
        sql = f"select max(`id`) as id from {tablename}"
        try:
            df = self.db.query(sql)
        except NotFound as e:
            return 0
        id = df['id'][0] if len(df) > 0 else 0
        return round(id)

    def _change_type_bq(self, tablename, fields):
        """tmpテーブルのnumber型と日付型を型変換する"""
        tmpl = (
            "drop table if exists " + tablename + "_tmp; "
            "create table " + tablename +
            "_tmp as select * except({0}), {1} from " +
            tablename + "_tmp2;"
        )
        excepts = ', '.join([
            f"`{k}`" for k, x in fields.items()
            if x['type'] in ('RECORD_NUMBER', 'NUMBER', 'DATETIME')
        ])
        casts = "safe_cast(`{0}` as {1}) as `{0}`"

        type_dic_bq = {
            'NUMBER': 'Numeric',
            'RECORD_NUMBER': 'Numeric',
            'DATETIME': 'Timestamp',
        }
        casts = ', '.join([
            casts.format(k, type_dic_bq[x['type']]) for k, x in fields.items()
            if x['type'] in ('RECORD_NUMBER', 'NUMBER', 'DATETIME')
        ])
        self.db.query_with_noreturn(tmpl.format(excepts, casts))

    def _rename_sql(self, table_name, old_names, new_name):
        """bigqueryのフィールド名を一括変換する"""
        cmd = f'alter table {table_name} '
        rename = [f'rename column `{x}` to `{y}`' for x, y in zip(old_names, new_name)]
        sql = cmd + ', '.join(rename) + ';'
        return sql

    def _select(self, app_name, where=None, fields=None, limit=None):
        """kintoneの情報を取得する"""
        info = self.apps[app_name]
        app = Kintone(info['api_token'], info['sub_domain'], info['app_id'])
        res = app.select_all(where, fields, hard_limit=limit)
        return pd.DataFrame(res), app.fields

    def _create_schema(self, df, fields):
        '''
        フィールド定義からbiqrueryのスキーマを生成する
        が、Parquetのチェックが厳しすぎて転送できないのでほぼStringで送ってbigquery側で型変換を行う
        '''
        type_dic = {
            'RADIO_BUTTON': 'bool',
            'DATE': 'DATE',
            'TIME': 'String',
            'UPDATED_TIME': 'TIMESTAMP',
            'DATETIME': 'TIMESTAMP',
            'CREATED_TIME': 'TIMESTAMP',
        }

        schema = [
            {
                'name': k, self.schema_type: type_dic[fields[k]['type']] if fields[k]['type'] in type_dic else 'String',
                'mode': self._mode(fields[k])
            }
            for k in df.columns
        ]
        return schema

    def _mode(self, f):
        ret = 'NULLABLE' if 'required' in f and (f['required'] is None or f['required'] == True) else ""
        return ret

    def _adjust_type(self, df, schema):
        '''
        データ型に基づいてdataframeの型を変換する
        '''
        type_conversion = {
            'bool': (bool, None),
            'BigNumeric': (float, None),
            'Integer': (int, None),
            'DATETIME': (None, pd.to_datetime),
            'TIMESTAMP': (None, pd.to_datetime),
            'DATE': ('datetime64[ns]', pd.to_datetime),
            'TIME': ('datetime64[ns]', pd.to_datetime),
            'String': (str, None)
        }

        for x, s in zip(df.columns, schema):
            dtype, converter = type_conversion.get(s[self.schema_type], (None, None))

            # データ型がマッピングに存在する場合、変換を実施
            if dtype is not None:
                df[x] = df[x].replace('', None)
                df[x] = converter(df[x].astype(dtype)) if converter else df[x].astype(dtype)
            elif converter is not None:
                df[x] = converter(df[x])

        return df

    def _clean_fieldname(self, columns):
        ''':
        bigqueryの規約に従って日本語フィールド名をクリーニングする.
        またid,revisionは_を取っておく。
        '''
        col = [x.translate(self.ng_fields) for x in columns]
        col = [x if x not in ('_id', '_revision') else x.replace('_', '') for x in col]
        return col



import boto3

class AWSTool:
    def __init__(self):
        pass

    def init_s3(self, region_name='ap-northeast-1'):
        self.s3 = boto3.resource('s3', region_name=region_name)
        return self

    def get_from_s3url(self, s3url):
        p = s3url.split('/')
        bucket, key = p[2], '/'.join(p[3:])
        binary = self.s3.Bucket(bucket).Object(key).get()['Body'].read()
        return binary

